import os
from pathlib import Path


def last_modified_file_in(dir: Path) -> str:
    """
    Return the path to the last modified file in `dir`.
    """
    all_files = Path(dir).glob(f"{dir}/*")
    return max(all_files, key=os.path.getctime)



import os
from pathlib import Path

import git
import modelator
import pytest
import typer
from copier import run_auto

from .. import chain, test
from ..reactor.reactor import generate_reactor

app = typer.Typer(
    add_completion=False,
    name="atomkraft",
    no_args_is_help=True,
    rich_markup_mode="rich",
)

GH_TEMPLATE = "gh:informalsystems/atomkraft"
GH_REVISION = "dev"


@app.command(
    no_args_is_help=True,
)
def init(
    name: Path = typer.Argument(..., help="Name of new directory", show_default=False)
):
    """
    Initialize new Atomkraft project in the given directory
    """
    try:
        git.Repo(os.getcwd(), search_parent_directories=True)
    except git.InvalidGitRepositoryError:
        git.Repo.init(name)
    run_auto(GH_TEMPLATE, name, vcs_ref=GH_REVISION)


app.add_typer(
    modelator.cli.app,
    name="model",
    short_help="Work with TLA+ models: load, typecheck, check invariants, or sample traces",
    no_args_is_help=True,
)
app.add_typer(
    chain.app,
    name="chain",
    help="Modify or control Cosmos-SDK chain",
    no_args_is_help=True,
)
app.add_typer(
    test.app,
    name="test",
    short_help="Test the blockchain based on abstract traces, either explicitly given or produced from models",
    no_args_is_help=True,
)


@app.command(
    no_args_is_help=True,
)
def smoke_test():
    """
    Run smoke tests for a Cosmos-SDK chain
    """
    pytest.main([".atomkraft/smoke_tests"])


# run with:
# poetry run atomkraft reactor --actions-list=act1,act2,act3 --variables-list=x,y,z --reactor-stub-file=path/to/the/file
# or
# poetry run atomkraft reactor --actions-list="act1, act2, act3" --variables-list="x, y, z" --reactor-stub-file="path/to/the/file"


@app.command(
    no_args_is_help=True,
)
def reactor(
    actions: str = typer.Option(
        ...,
        help="trace actions for which to create reactor stub functions",
        show_default=False,
    ),
    variables: str = typer.Option(
        ...,
        help="state variables to use as parameters for reactor stub functions.",
        show_default=False,
    ),
    path: typer.FileTextWrite = typer.Option(
        os.path.abspath("reactors/reactor.py"),
        help="path where to create the reactor stub",
    ),
):
    """
    Generate a reactor stub used to interpret test traces
    """
    actions = [act.strip() for act in actions.split(",")]
    variables = [var.strip() for var in variables.split(",")]

    if Path(path.name).is_file():
        typer.confirm(
            "The stub file already exists and it will be overwritten. "
            "Are you sure you want to continue?",
            abort=True,
        )

    generate_reactor(actions, variables, stub_file_path=path.name)



import pytest

from ..utils.project import project_root
from .testnet import Testnet


@pytest.fixture
def testnet():
    testnet = Testnet.load_toml(f"{project_root()}/chain.toml")

    yield testnet



from pathlib import Path
from typing import List

from atomkraft.config.atomkraft_config import AtomkraftConfig
from atomkraft.config.model_config import ModelConfig
from atomkraft.utils.filesystem import last_modified_file_in
from modelator.cli.model_config_file import load_config_file
from modelator.itf import ITF
from modelator.Model import Model
from modelator.ModelResult import ModelResult


def query_configs(key: str) -> str:
    with AtomkraftConfig() as atomkraft_config:
        with ModelConfig() as model_config:
            try:
                return model_config.query(key) or atomkraft_config.query(key)
            except KeyError:
                raise FileNotFoundError


def generate_traces(
    model_config_path=None, model_path=None, sample_operators=[]
) -> ModelResult:
    """
    Call Modelator to get samples of the given model in `model_path`. Return the
    result of running the checker as a ModelResult.

    If a `model_config_path` is provided, use that to load the model path.
    Otherwise, use `model_path` and `sample_operators`.

    Raise an exception when a required parameter is absent, or when they are not
    available via defaults, or when the parameters are invalid.
    """
    init = "Init"
    next = "Next"
    if model_config_path:
        model_config = load_config_file(model_config_path)
        model_path = model_config["model_path"]
        init = model_config["init"] or init
        next = model_config["next"] or next
        sample_operators = list(set(model_config["examples"] + sample_operators))

    if not model_path:
        raise ValueError("No model path given.")

    if not Path(model_path).is_file():
        raise FileNotFoundError(f"File with model not found: {model_path}")

    model = Model.parse_file(model_path, init, next)
    return model.sample(examples=sample_operators)


def last_modified_trace_path() -> str:
    """
    Return the path to the last modified trace file in the trace directory.
    """
    try:
        traces_dir = query_configs("traces_dir")
    except FileNotFoundError:
        raise FileNotFoundError(
            "No trace directory set in neither model or configuration file."
        )

    if not Path(traces_dir).is_dir():
        raise FileNotFoundError(f"Directory does not exist: {traces_dir}")

    return last_modified_file_in(traces_dir)


def get_trace(trace_path=None) -> List["ITF"]:
    """
    Return a trace represented as an ITF class from Modelator. If `trace_path`
    is absent, return the last trace generated with the `model` command.
    """
    if not trace_path:
        trace_path = last_modified_trace_path()

    if not Path(trace_path).is_file():
        raise FileNotFoundError(f"Path {trace_path} is not a file")

    print(f"Retrieving trace from: {trace_path}")
    return ITF.from_itf_json(trace_path)



import glob
import time
from pathlib import Path
from typing import Dict

import click
import numpy as np
import tabulate
import tomlkit

from .node import Account, Coin, ConfigPort, Node
from .utils import get_free_ports, update_port


class Testnet:
    def __init__(
        self,
        chain_id,
        n_validator=2,
        n_account=0,
        binary="gaiad",
        denom="uatom",
        prefix="cosmos",
        coin_type=118,
        genesis_config: Dict = {},
        node_config: Dict = {},
        account_balance: int = 10**27,
        validator_balance: int = 10**21,
        overwrite=True,
        keep=True,
        verbose=False,
        data_dir=None,
    ):
        self.chain_id = chain_id
        self.n_validator = n_validator
        self.n_account = n_account
        self.binary = binary
        self.denom = denom
        self.prefix = prefix
        self.coin_type = coin_type
        self.genesis_config = genesis_config
        self.node_config = node_config
        self.account_balance = account_balance
        self.validator_balance = validator_balance
        self.overwrite = overwrite
        self.keep = keep
        self.verbose = verbose
        self.data_dir = Path(".atomkraft/nodes") if data_dir is None else data_dir

    @staticmethod
    def load_toml(path: str):
        with open(path) as f:
            data = tomlkit.load(f)

        return Testnet(
            data["chain_id"],
            data["n_validator"],
            data["n_account"],
            data["binary"],
            data["denom"],
            data["prefix"],
            data["coin_type"],
            data["config"]["genesis"],
            data["config"]["node"],
        )

    @staticmethod
    def ports() -> Dict[str, ConfigPort]:
        data = {}

        # config.toml
        data["p2p"] = ConfigPort("P2P", "config/config.toml", "p2p.laddr")
        # data["p2p_ext"] = ConfigPort("P2P External", "config/config.toml", "p2p.external_address")
        data["abci"] = ConfigPort("ABCI", "config/config.toml", "proxy_app")
        data["pprof_laddr"] = ConfigPort(
            "PPROF", "config/config.toml", "rpc.pprof_laddr"
        )
        data["rpc"] = ConfigPort("RPC", "config/config.toml", "rpc.laddr")
        data["prometheus"] = ConfigPort(
            "Prometheus", "config/config.toml", "instrumentation.prometheus_listen_addr"
        )

        # app.toml
        data["lcd"] = ConfigPort("LCD", "config/app.toml", "api.address")
        data["grpc"] = ConfigPort("gRPC", "config/app.toml", "grpc.address")
        data["grpc-web"] = ConfigPort("gRPC", "config/app.toml", "grpc-web.address")
        # dict["rosetta"] = ConfigPort("Rosetta", "config/app.toml", "rosetta.address")
        return data

    def get_validator_port(self, validator_id, port_type):
        return self.validator_nodes[validator_id].get_port(self.ports()[port_type])

    def prepare(self):
        self.data_dir.mkdir(parents=True, exist_ok=True)
        self.validator_nodes = [
            Node(
                f"node-{i}",
                self.chain_id,
                f"{self.data_dir}/node-{i}",
                overwrite=self.overwrite,
                keep=self.keep,
                binary=self.binary,
                denom=self.denom,
                prefix=self.prefix,
            )
            for i in range(self.n_validator)
        ]
        self.validators = [Account(f"validator-{i}") for i in range(self.n_validator)]
        self.accounts = [Account(f"account-{i}") for i in range(self.n_account)]

        for node in self.validator_nodes:
            node.init()

        for (k, v) in self.genesis_config.items():
            self.validator_nodes[0].set("config/genesis.json", v, k)

        for validator in self.validators:
            self.validator_nodes[0].add_account(
                Coin(self.account_balance, self.denom), validator
            )

        for account in self.accounts:
            self.validator_nodes[0].add_account(
                Coin(self.account_balance, self.denom), account
            )

        for i in range(1, len(self.validator_nodes)):
            self.validator_nodes[i].copy_genesis_from(self.validator_nodes[0])

        # very hacky
        all_ports = np.array(
            get_free_ports(len(self.ports()) * (self.n_validator - 1))
        ).reshape((-1, len(self.ports())))

        all_port_data = []

        for (i, node) in enumerate(self.validator_nodes):
            for (config_file, configs) in self.node_config.items():
                for (k, v) in configs.items():
                    if isinstance(v, str):
                        # TODO: avoid tomlkit.items.str being a list
                        node.set(f"config/{config_file}.toml", str(v), k)
                    else:
                        node.set(f"config/{config_file}.toml", v, k)

            if i > 0:
                ports = all_ports[i - 1]
                for (j, (key, e_port)) in enumerate(self.ports().items()):
                    node.update(
                        e_port.config_file,
                        lambda x: update_port(x, ports[j]),
                        e_port.property_path,
                    )

            port_data = [node.moniker]
            for (j, e_port) in enumerate(self.ports().values()):
                port_data.append(node.get(e_port.config_file, e_port.property_path))
            all_port_data.append(port_data)

        if self.verbose:
            print(
                tabulate.tabulate(
                    all_port_data,
                    headers=["Moniker"] + [e.title for e in self.ports().values()],
                )
            )

            print(
                tabulate.tabulate(
                    [[validator.address(self.prefix)] for validator in self.validators],
                    headers=["Validators"],
                )
            )

            print(
                tabulate.tabulate(
                    [[account.address(self.prefix)] for account in self.accounts],
                    headers=["Accounts"],
                )
            )

        for (i, node) in enumerate(self.validator_nodes):
            node.add_key(self.validators[i])
            p2p = self.ports()["p2p"]
            node.add_validator(
                Coin(self.validator_balance, self.denom), self.validators[i]
            )

            if i > 0:
                # because this
                # https://github.com/cosmos/cosmos-sdk/blob/88ee7fb2e9303f43c52bd32410901841cad491fb/x/staking/client/cli/tx.go#L599
                gentx_file = glob.glob(f"{node.home_dir}/config/gentx/*json")[0]
                gentx_file = gentx_file.removeprefix(node.home_dir)
                node_p2p = node.get(p2p.config_file, p2p.property_path).rsplit(
                    ":", maxsplit=1
                )[-1]
                node.update(gentx_file, lambda x: update_port(x, node_p2p), "body.memo")
                node.sign(self.validators[i], f"{node.home_dir}/{gentx_file}")

        for i in range(1, len(self.validator_nodes)):
            for j in range(i):
                self.validator_nodes[i].copy_gentx_from(self.validator_nodes[j])
                self.validator_nodes[j].copy_gentx_from(self.validator_nodes[i])

        for node in self.validator_nodes:
            node.collect_gentx()

    def spinup(self):
        for node in self.validator_nodes:
            node.start()

    def oneshot(self):
        self.prepare()
        self.spinup()

    def teardown(self):
        for node in self.validator_nodes:
            node.close()


@click.command()
def testnet(
    chain_id: str,
    validator: int,
    account: int,
    binary: str,
    denom: str,
    prefix: str,
    coin_type: int,
    overwrite: bool,
    keep: bool,
    data_dir: Path,
):
    coin_type = 118

    genesis_config = {
        "app_state.gov.voting_params.voting_period": "600s",
        "app_state.mint.minter.inflation": "0.300000000000000000",
    }

    node_config = {
        "app": {
            "api.enable": True,
            "api.swagger": True,
            "api.enabled-unsafe-cors": True,
            "minimum-gas-prices": f"0.10{denom}",
            "rosetta.enable": False,
        },
        "config": {
            "instrumentation.prometheus": False,
            "p2p.addr_book_strict": False,
            "p2p.allow_duplicate_ip": True,
        },
    }

    net = Testnet(
        chain_id,
        n_validator=validator,
        n_account=account,
        binary=binary,
        denom=denom,
        prefix=prefix,
        coin_type=coin_type,
        genesis_config=genesis_config,
        node_config=node_config,
        account_balance=10**26,
        validator_balance=10**16,
        overwrite=overwrite,
        keep=keep,
        data_dir=data_dir,
    )

    net.oneshot()
    try:
        time.sleep(600)
    except KeyboardInterrupt:
        print("\ntear-down!")



import os
from pathlib import Path


def project_root():
    cwd = Path(os.getcwd())
    while cwd != cwd.parent:
        if (
            (cwd / "pyproject.toml").exists()
            and (cwd / "atomkraft.toml").exists()
            and (cwd / ".atomkraft" / "config.toml").exists()
        ):
            return cwd
        cwd = cwd.parent



import os.path
import shutil
from datetime import datetime
from os import PathLike
from pathlib import Path
from typing import List

import pytest
from atomkraft.utils.project import project_root

from ..reactor.reactor import get_reactor

TRACE_TEST_STUB = """
from modelator.pytest.decorators import itf

pytest_plugins = "{0}"

@itf("{1}", keypath="{2}")
def test_trace():
    print("Successfully executed trace {1}")
"""


def copy(src_paths: List[Path], dst_path: Path):
    for src in src_paths:
        if src.is_dir():
            shutil.copytree(src, dst_path / src.name)
        else:
            shutil.copy2(src, dst_path)


def test_trace(trace: PathLike, reactor: PathLike, keypath: str, verbose: bool):
    """
    Test blockchain by running one trace
    """

    if reactor is None:
        reactor = get_reactor()

    root = project_root()
    if not root:
        raise RuntimeError(
            "could not find Atomkraft project: are you in the right directory?"
        )

    tests = root / "tests"
    tests.mkdir(exist_ok=True)

    timestamp = datetime.now().isoformat(timespec="milliseconds")
    test_name = f"test_{str(trace)}_{timestamp}"
    test_name = (
        test_name.replace("/", "_")
        .replace(".", "_")
        .replace(":", "_")
        .replace("-", "_")
    )
    test_path = os.path.join(tests, f"{test_name}.py")
    with open(test_path, "w") as test:
        print(f"Writing {test_name} ...")
        test.write(
            TRACE_TEST_STUB.format(
                str(reactor).replace("/", ".").removesuffix(".py"), trace, keypath
            )
        )
    print(f"Executing {test_name} ...")

    report_dir = root / "reports" / test_name
    report_dir.mkdir(parents=True)

    logging_file = report_dir / "log.txt"

    pytest_report_file = report_dir / "report.jsonl"

    pytest_args = [
        "--log-file-level=INFO",
        f"--log-file={logging_file}",
        f"--report-log={pytest_report_file}",
    ]

    if verbose:
        pytest_args.append("-rP")
        pytest_args.append("--log-cli-level=INFO")

    pytest.main(pytest_args + [test_path])

    copy([Path(trace), root / ".atomkraft" / "nodes"], report_dir)

    print(f"Test data is saved at {report_dir}")



import json
import time
from pathlib import Path
from typing import Optional

import tomlkit
import typer
from atomkraft.utils import project, query, update

from .node import Account, Coin, ConfigPort, Node
from .testnet import Testnet

__all__ = ["Testnet", "Account", "Node", "ConfigPort", "Coin"]


app = typer.Typer()


@app.command()
def config(
    property_path: str = typer.Argument(
        ..., help="Nested keys of a config value, joined by `.`", show_default=False
    ),
    value: Optional[str] = typer.Argument(
        None, help="Update old value with provided value"
    ),
):
    """
    Query or update chain configuration
    """
    if value is None:
        with open(f"{project.project_root()}/chain.toml") as f:
            data = tomlkit.load(f)
        print(json.dumps({property_path: query(data, property_path)}, indent=2))
    else:
        try:
            value = eval(value)
        except (SyntaxError, NameError):
            pass
        with open(f"{project.project_root()}/chain.toml") as f:
            data = update(tomlkit.load(f), property_path, value)
        with open(f"{project.project_root()}/chain.toml", "w") as f:
            tomlkit.dump(data, f)


@app.command()
def testnet(
    silent: bool = typer.Option(False, help="Silent mode. Print no output"),
    config: Optional[Path] = typer.Option(None, help="Path to chain.toml"),
):
    """
    Run a testnet in background
    """
    if config is None:
        testnet = Testnet.load_toml(f"{project.project_root()}/chain.toml")
    else:
        testnet = Testnet.load_toml(config)
    testnet.verbose = not silent

    testnet.oneshot()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("\ntear-down!")



import glob
import json
import os
import shutil
from dataclasses import dataclass
from subprocess import PIPE, Popen

import bip_utils
import hdwallet
import numpy as np
import tomlkit
from hdwallet.symbols import ATOM

from .. import utils


@dataclass
class ConfigPort:
    title: str
    config_file: str
    property_path: str


class Client:
    def __init__(self, home_dir, binary="gaiad"):
        self.home_dir = home_dir
        self.binary = binary


@dataclass
class Coin:
    denom: str
    amount: int
    expoent: int

    def __init__(self, amount, denom, exponent=6):
        self.amount = amount
        self.denom = denom
        self.expoent = exponent

    def __repr__(self):
        return f"{self.amount}{self.denom}"


class Account:
    def __init__(self, name, seed=None, strength=128):
        if strength not in [128, 160, 192, 224, 256]:
            raise ValueError(
                f"Strength should be one of the following [128, 160, 192, 224, 256], but it is not {strength}."
            )
        self.name = name
        if seed is None:
            seed = self.name
        self.entropy = (
            np.random.default_rng(list(seed.encode())).bytes(strength // 8).hex()
        )
        self.wallet = hdwallet.BIP44HDWallet(symbol=ATOM).from_entropy(
            entropy=self.entropy, language="english", passphrase=""
        )

    @property
    def mnemonic(self):
        return self.wallet.mnemonic()

    def address(self, prefix):
        return self.bech32_address(prefix)

    def validator_address(self, prefix):
        return self.bech32_address(f"{prefix}valoper")

    def bech32_address(self, prefix):
        return bip_utils.Bech32Encoder.Encode(prefix, bytes.fromhex(self.wallet.hash()))

    def __repr__(self):
        return json.dumps(self.wallet.dumps(), indent=2)


class Node:
    def __init__(
        self,
        moniker,
        chain_id,
        home_dir,
        *,
        overwrite=False,
        keep=False,
        binary="gaiad",
        denom="uatom",
        prefix="cosmos",
    ):
        self.moniker = moniker
        self.chain_id = chain_id
        self.home_dir = home_dir
        self.binary = binary
        self.denom = denom
        self.prefix = prefix
        self.overwrite = overwrite
        self.keep = keep
        self._popen = None
        self._stdout = None
        self._stderr = None

    def init(self):
        if self.overwrite and os.path.exists(self.home_dir):
            shutil.rmtree(self.home_dir)
        args = f"init {self.moniker} --chain-id {self.chain_id}".split()
        _, data = self._execute(args, stderr=PIPE)
        return json.loads(data.decode())

    def add_key(self, account: Account):
        argstr = (
            f"keys add {account.name} --recover --keyring-backend test --output json"
        )
        stdout, stderr = self._execute(
            argstr.split(),
            stdin=f"{account.wallet.mnemonic()}\n".encode(),
            stdout=PIPE,
            stderr=PIPE,
        )
        if stdout:
            return json.loads(stdout.decode())
        return json.loads(stderr.decode())

    def add_account(self, coin: Coin, account: Account):
        argstr = f"add-genesis-account {account.address(self.prefix)} {coin} --keyring-backend test --output json"
        self._execute(argstr.split())

    def add_validator(self, coin: Coin, account: Account):
        argstr = f"gentx {account.name} {coin} --keyring-backend test --chain-id {self.chain_id} --output json"
        self._execute(argstr.split(), stderr=PIPE)

    def collect_gentx(self):
        argstr = "collect-gentxs"
        _, data = self._execute(argstr.split(), stderr=PIPE)
        return json.loads(data.decode())

    def copy_gentx_from(self, other: "Node"):
        for file in glob.glob(f"{other.home_dir}/config/gentx/*.json"):
            shutil.copy(file, f"{self.home_dir}/config/gentx")

    def copy_genesis_from(self, other: "Node"):
        for file in glob.glob(f"{other.home_dir}/config/genesis.json"):
            shutil.copy(file, f"{self.home_dir}/config")

    def get(self, path, property_path=None):
        with open(f"{self.home_dir}/{path}", encoding="utf-8") as f:
            match os.path.splitext(path)[-1]:
                case ".json":
                    data = json.load(f)
                case ".toml":
                    data = tomlkit.load(f)
                case _:
                    raise RuntimeError(f"Unexpected file {path}")
        return utils.query(data, property_path)

    def get_port(self, port_config: ConfigPort):
        return self.get(
            port_config.config_file,
            port_config.property_path,
        )

    def set(self, path, value, property_path=None):
        if property_path is not None:
            with open(f"{self.home_dir}/{path}", encoding="utf-8") as f:
                match os.path.splitext(path)[-1]:
                    case ".json":
                        main_data = json.load(f)
                    case ".toml":
                        main_data = tomlkit.load(f)
                    case _:
                        raise RuntimeError(f"Unexpected file {path}")
            main_data = utils.update(main_data, property_path, value)
        else:
            main_data = value
        with open(f"{self.home_dir}/{path}", "w", encoding="utf-8") as f:
            match os.path.splitext(path)[-1]:
                case ".json":
                    json.dump(main_data, f)
                case ".toml":
                    tomlkit.dump(main_data, f)
                case _:
                    raise RuntimeError(f"Unexpected file {path}")

    def update(self, path, function, property_path=None):
        self.set(path, function(self.get(path, property_path)), property_path)

    def _execute(self, args, *, stdin: bytes | None = None, stdout=None, stderr=None):
        final_args = f"{self.binary} --home {self.home_dir}".split() + args
        # print(" ".join(final_args))
        stdin_pipe = None if stdin is None else PIPE
        with Popen(final_args, stdin=stdin_pipe, stdout=stdout, stderr=stderr) as p:
            out, err = p.communicate(input=stdin)
            rt = p.wait()
            if rt != 0:
                raise RuntimeError(f"Non-zero return code {rt}\n{err.decode()}")
            return (out, err)

    def __enter__(self):
        return self

    def sign(self, account, tx_file: str):
        argstr = f"tx sign {tx_file} --output-document {tx_file} --chain-id {self.chain_id} --overwrite --offline --sequence 0 --account-number 0 --from {account.name} --keyring-backend test --output json"
        self._execute(argstr.split())

    def start(self):
        self._stdout = open(f"{self.home_dir}/stdout.txt", "w", encoding="utf-8")
        self._stderr = open(f"{self.home_dir}/stderr.txt", "w", encoding="utf-8")
        args = f"{self.binary} start --home {self.home_dir}".split()
        self._popen = Popen(args, stdout=self._stdout, stderr=self._stderr)

    def close(self):
        if self._popen:
            self._popen.terminate()
            self._stdout.close()
            self._stderr.close()
            self._popen = None
            self._stdout = None
            self._stderr = None
        if not self.keep and os.path.isdir(self.home_dir):
            shutil.rmtree(self.home_dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()



import tomlkit


class ConfigFile(object):
    def __init__(self, path: str):
        self.data = {}
        self.path = path

    def __enter__(self):
        self.fd = open(self.path, "w+")
        self.data = tomlkit.load(self.fd)
        return self

    def __exit__(self, type, value, traceback):
        self.fd.close()

    def query(self, key) -> str:
        if key not in self.data:
            raise KeyError
        return self.data[key]

    def store(self, key, value):
        self.data[key] = value
        tomlkit.dump(self.data, self.fd)



import ast
from os import PathLike
from typing import List

from atomkraft.config.atomkraft_config import AtomkraftConfig
from caseconverter import snakecase

from . import constants, utils
from .step_functions_visitor import StepFunctionsVisitor


def check_reactor(trace: PathLike, reactor=None) -> bool:
    """
    returns true if each action appearing in `trace` is matched with a function in `reactor`
    """
    if reactor is None:
        reactor = get_reactor()

    with open(reactor) as f_reactor:
        root_node = ast.parse(f_reactor.read())
    v = StepFunctionsVisitor()
    v.visit(root_node)

    keypath = utils.get_keypath(root_node)
    try:
        all_trace_actions = utils.get_all_trace_actions(trace, keypath)
    except Exception:
        raise ValueError(
            f"Keypath {keypath}, which is set in {reactor} is not present in the trace {trace}"
        )

    if constants.ALL_ENCOMPASSING_STEP in v.step_functions:
        return True
    else:
        return all_trace_actions.issubset(v.step_functions)


def get_reactor() -> PathLike:
    """
    returns the path to the current reactor from the internal config
    """
    with AtomkraftConfig() as config:
        try:
            return config.query(constants.REACTOR_CONFIG_KEY)
        except KeyError:
            raise RuntimeError(
                "Could not find default reactor; have you ran `atomkraft reactor`?"
            )


def generate_reactor(
    actions_list: List[str],
    variables_list: List[str],
    stub_file_path: PathLike,
    keypath: str = "action",
) -> PathLike:

    file_comment = _file_comment()
    imports_stub = _imports_stub()

    state_stub = _state_stub()
    actions_stub = "\n".join(
        [
            _action_stub(action_name=act, variables=variables_list)
            for act in actions_list
        ]
    )
    keypath_stub = _keypath_stub(keypath)
    with open(stub_file_path, "w") as f:
        f.write(file_comment)
        f.write(imports_stub)
        f.write(keypath_stub)
        f.write(state_stub)
        f.write(actions_stub)

    with AtomkraftConfig() as config:
        config.store(constants.REACTOR_CONFIG_KEY, stub_file_path)

    return stub_file_path


def _file_comment():
    return """'''
The reactor file connects a test scenario described by a trace
(obtained from a model or written by hand) with the actual execution
of the test scenario on the testnet.

It contains one @step decorated function per each action appearing in the trace:
those function implement changes to the blockchain corresponding to the
abstract action from the trace.

All step functions receive the following arguments:
    testnet: a `Testnet` object on which blockchain transactions can be
             executed.
    state: additional logical state, as defined the the `state()`
           function in this file.
    action: object from the trace which corresponds to the parameters
            of the taken step.
'''
"""


def _keypath_stub(keypath):
    stub = f"""

{constants.KEYPATH_VAR} = {repr(keypath)}
"""
    return stub


def _action_description_comment(action_name, variables):
    if len(variables) == 0:
        variables_sentence = ""
    elif len(variables) == 1:
        variables_sentence = f"It additionally has access to the model (trace) state variable `{variables[0]}`."
    elif len(variables) == 2:
        variables_sentence = f"It additionally has access to the model (trace) state variables `{variables[0]}` and `{variables[1]}`."
    else:
        vars_string = "".join([f"\n\t\t-`{v}`" for v in variables])
        variables_sentence = f"It additionally has access to the model (trace) state variables: {vars_string}."
    return f"""'''
    Implements the effects of the step {repr(action_name)}
    on blockchain `testnet` and state `state`.
    {variables_sentence}
    '''
    """


def _action_stub(action_name: str, variables: List[str]):
    stub = f"""


@step({repr(action_name)})
def {snakecase(action_name)}(testnet, state, {", ".join(variables)}):
    {_action_description_comment(action_name, variables)}
    #TODO: replace the printing stub with the effects of the action `{action_name}`
    print("Step: {action_name}")
"""
    return stub


def _state_stub():
    stub = """


@pytest.fixture
def state():
    '''
    Defines any additional logical state (beyond the state of the chain)
    that needs to be maintained throughout the execution. This state
    will be passed as an argument to @step functions.
    '''

    #TODO: replace the empty stub object with a different state object
    # if necessary
    return {}
"""
    return stub


def _imports_stub():
    stub = """
import pytest
from modelator.pytest.decorators import step
"""
    return stub



# a key for the reactor path inside internal config
REACTOR_CONFIG_KEY = "reactor"

# a special value used to indicate that there is a @step-decorated function that matches all actions
# (because it has no arguments)
ALL_ENCOMPASSING_STEP = object()

# a variable created under generated `reactor.py` file which stores the value of the keypath
KEYPATH_VAR = "keypath"



import copy
from typing import Any, Dict


def query(
    data: Dict[Any, Any],
    property_path: str | None,
) -> Any:
    if property_path:
        keys = property_path.split(".")
        for key in keys:
            match data:
                case dict():
                    try:
                        data = data[key]
                    except KeyError:
                        try:
                            data = data[key.replace("-", "_")]
                        except KeyError:
                            data = data[key.replace("_", "-")]
                        except Exception as e:
                            raise e
                case list():
                    data = data[int(key)]
    return data


def merge(old: Any, new: Any):
    if isinstance(old, Dict):
        old = copy.deepcopy(old)
        for (key, val) in new.items():
            if key in old and isinstance(val, dict):
                old[key] = merge(old[key], val)
            else:
                old[key] = val
        return old
    else:
        return new


def update(data_root: Dict[Any, Any], property_path: str | None, value: Any) -> Any:
    if property_path:
        data = data_root
        keys = property_path.split(".")
        for key in keys[:-1]:
            match data:
                case dict():
                    try:
                        data = data[key]
                    except KeyError:
                        try:
                            data = data[key.replace("-", "_")]
                        except KeyError:
                            data = data[key.replace("_", "-")]
                case list():
                    data = data[int(key)]
        match data:
            case dict():
                key = keys[-1]
                try:
                    data[key] = merge(data[key], value)
                except KeyError:
                    try:
                        cust_key = key.replace("-", "_")
                        data[cust_key] = merge(data[cust_key], value)
                    except KeyError:
                        cust_key = key.replace("_", "-")
                        data[cust_key] = merge(data[cust_key], value)
            case list():
                cust_key = int(keys[-1])
                data[cust_key] = merge(data[cust_key], value)
    else:
        data_root = merge(data_root, value)
    return data_root



import asyncio
import json
import socket

import websockets
from jsonrpcclient import responses
from jsonrpcclient.requests import request_json


def get_free_ports(n):
    socks = []
    ports = []
    for _ in range(n):
        sock = socket.socket()
        sock.bind(("", 0))
        ports.append(sock.getsockname()[1])
        socks.append(sock)
    for sock in socks:
        sock.close()
    return ports


def update_port(old, new):
    old_port = old.split(":")[-1]
    return old.replace(old_port, str(new))


class TmEventSubscribe:
    def __init__(self, params):
        self.params = params

    async def _subscribe(self, params):
        self.websocket = await websockets.connect("ws://0.0.0.0:26657/websocket")
        self.query = " AND ".join(
            f"{key} = '{value}'" for (key, value) in params.items()
        )
        req = request_json("subscribe", params={"query": self.query})
        await self.websocket.send(req)

    def __enter__(self):
        self.loop = asyncio.new_event_loop()
        self.loop.run_until_complete(self._subscribe(self.params))
        return self

    async def _wait(self):
        while True:
            msg = responses.to_result(json.loads(await self.websocket.recv()))
            if "query" in msg.result and msg.result["query"] == self.query:
                break

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.loop.run_until_complete(self._wait())
        self.loop.run_until_complete(self.websocket.close())
        self.loop.close()



import os
from typing import Optional

from atomkraft.config.config_file import ConfigFile
from atomkraft.utils.project import project_root

ATOMKRAFT_INTERNAL_DIR = ".atomkraft"

# the file that contains internal configurations for atomkraft
ATOMKRAFT_INTERNAL_CONFIG = "config.toml"


class AtomkraftConfig(ConfigFile):
    def __init__(self, path: Optional[str] = None):
        if not path:
            if "PROJECT_ROOT" in os.environ:
                root = os.environ["PROJECT_ROOT"]
            else:
                root = project_root()
            path = os.path.join(
                root,
                ATOMKRAFT_INTERNAL_DIR,
                ATOMKRAFT_INTERNAL_CONFIG,
            )
        super().__init__(path)



import ast
from typing import Set

from . import constants


class StepFunctionsVisitor(ast.NodeVisitor):
    """
    This visitor will upon visiting a function definition explore if this function is decorated
    by @step (in method visit_FunctionDef). All other nodes will be simply traversed.
    """

    def __init__(self) -> None:
        self.step_functions: Set[str] = set()
        super().__init__()

    def visit_FunctionDef(self, node: ast.FunctionDef) -> bool:
        for dec in node.decorator_list:
            if StepFunctionsVisitor._is_step(dec):
                self.step_functions.add(StepFunctionsVisitor.step_name(dec))

    @classmethod
    def _is_step(cls, dec) -> bool:

        # only accepts decorators of the form @step(...)
        if isinstance(dec, ast.Call):
            if dec.func.id == "step":
                return True
        return False

    @classmethod
    def step_name(cls, dec):
        # get all the arguments of the decorator `dec`
        all_args = [a.value for a in dec.args]

        if len(all_args) == 0:
            # if it comes with no arguments, this is a decorator that can be applied to any action
            return constants.ALL_ENCOMPASSING_STEP
        elif len(all_args) == 1:
            return all_args[0]
        else:
            # step decorators should never receive more than one argument. Raising exception if that happens
            raise ValueError(
                "Step decorators accept only one argument: a name of the step to be matched by the decorated function"
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['atomkraft',
 'atomkraft.chain',
 'atomkraft.cli',
 'atomkraft.config',
 'atomkraft.model',
 'atomkraft.reactor',
 'atomkraft.test',
 'atomkraft.utils']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0',
 'bip-utils>=2.3.0,<3.0.0',
 'case-converter>=1.1.0,<2.0.0',
 'copier>=6.1.0,<7.0.0',
 'hdwallet>=2.1.1,<3.0.0',
 'jsonrpcclient>=4.0.2,<5.0.0',
 'modelator>=0.6.0,<0.7.0',
 'numpy>=1.22.4,<2.0.0',
 'pytest-reportlog>=0.1.2,<0.2.0',
 'pytest>=7.1.2,<8.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'tomlkit>=0.11.1,<0.12.0',
 'typer[all]>=0.6.1,<0.7.0',
 'websockets>=10.3,<11.0']

entry_points = \
{'console_scripts': ['atomkraft = atomkraft.cli:app'],
 'pytest11': ['atomkraft-chain = atomkraft.chain.pytest']}

setup_kwargs = {
    'name': 'atomkraft',
    'version': '0.0.8',
    'description': 'Testing for Cosmos Blockchains',
    'long_description': "# Atomkraft: E2E testing for Cosmos blockchains\n\nBelow we describe what Atomkraft is about, and explain the concepts behind the tool. In case you would like to skip that, and jump directly into action, please read our [Installation guide](./INSTALLATION.md) to install the tool; afterwards it's worth following either our [Cosmos SDK Token Transfer tutorial](examples/cosmos-sdk/transfer/transfer.md), or [CosmWasm tutorial](examples/cosmwasm/counter/README.md).\n\nWe cover the following topics in this file:\n\n- [Atomkraft Introduction](#atomkraft-introduction)\n- [Conceptual overview](#conceptual-overview)\n  - [Atomkraft project](#atomkraft-project)\n  - [Local testnets](#local-testnets)\n  - [Traces and reactors](#traces-and-reactors)\n  - [Generating traces from TLA+ models](#generating-traces-from-tla-models)\n  - [Running the tests against testnet](#running-the-tests-against-testnet)\n- [What's next: Atomkraft's immediate future](#whats-next-atomkrafts-immediate-future)\n\n## Atomkraft Introduction\n\nThe [Cosmos Network](https://cosmos.network) of [IBC](https://ibcprotocol.org)-connected blockchains is growing tremendously fast. There is one aspect though, which has not been fully addressed yet, namely **quality assurance**: how do we make sure that Cosmos-based blockchains are secure, and don't contain security issues that pose hazards for user funds?\n\n**Atomkraft** is our answer to that question: it allows to generate and execute massive end-to-end (E2E) test suites for Cosmos SDK based blockchains. When designing the tool, we keep two main categories of users for it: **security auditors**, and **Cosmos SDK blockchain developers**. Those roles are abstract and non-exclusive, and they serve only to summarize some important concerns wrt. the deployed blockchains. Is Atomkraft the right tool for you? Below are some possible hints.\n\nAtomkraft is likely to benefit you as a _Security auditor_ if:\n\n- your blockchain (or the new version of it) is about to be launched, but you are unsure whether incentivized testnets allowed you to discover all critical vulnerabilities;\n- you are on tight time schedule for the new release, and you feel that the implementation has been rushed a little, so you are unsure that all corner cases of the new functionality has been covered.\n\nAtomkraft is likely to benefit you as a _Cosmos SDK developer_ if:\n\n- the new functionality you've implemented interacts in non-trivial ways with other important modules (e.g. with [bank](https://docs.cosmos.network/master/modules/bank/), [staking](https://docs.cosmos.network/master/modules/staking/), [authz](https://docs.cosmos.network/master/modules/authz/), etc.), and you are unsure whether some important invariants of those modules are preserved;\n- you want to grow and maintain a regression test suite for the blockchain modules you are developing, to make sure of their correctness as your blockchain evolves;\n- you want to automate quality assurance for your blockchain modules, and integrate fast E2E testing solution that's executed on every PR.\n\n**Key Atomkraft features** that allow to address the above concerns:\n\n- Push-button, but fully customizable, automation of local testnet creation\n- Clean, fully customizable trajectory from test case generation to its execution on the local testnet\n- Support for automatic generation of massive test suites from compact TLA+ models\n- Easy execution of test cases generated via other means (e.g. manual, BDD, PBT)\n- Anytime ready-to-integrate regression test suite in the form of a standard Pytest project\n- Ready-to-execute standard test suites for important Cosmos SDK modules (coming soon!)\n- Generation of reports and dashboards for presentation and analysis of testing results (coming soon!)\n\n## Conceptual overview\n\nIn itself, Atomkraft is nothing more than a command-line application, which is as easy to obtain for your system as executing `pip install atomkraft`; please consult the detailed [Installation Instructions](INSTALLATION.md) if needed. At the top-level, Atomkraft provides you the commands illustrated in the diagram below.\n\n![Atomkraft users](docs/images/atomkraft-users.svg)\n\n### Atomkraft project\n\nWorking with Atomkraft takes place in the scope of an _Atomkraft project_, which is initialized via `atomkraft init` command. The command will setup a new project in a given directory and install all necessary software for it, all fully automatically.\n\nThe good news is that an Atomkraft project is simultaneously a [Pytest](https://docs.pytest.org/) project! At any point in time you can execute `pytest` inside your Atomkraft project, and this command will rerun all tests you've created in the scope of this project. So, when you've executed `atomkraft init`, you've also started to create your regression test suite; congratulations!\n\nMore than that, an Atomkraft project is also a [Python Poetry](https://python-poetry.org) project: you can treat it just as a normal Python project, and add your custom Python tests there.\n\nAt the top level, an Atomkraft project contains the following:\n\n- Folders that hold the corresponding artifacts:\n  - `models` is meant to keep your TLA+ models, if you are willing to generate test cases from formal models;\n  - `traces` contain abstract test cases, represented in the form of [ITF traces](https://apalache.informal.systems/docs/adr/015adr-trace.html) (a JSON encoding of a sequence of test case steps);\n  - `reactors` hold Python functions that interpret ITF traces, and map trace steps to concrete blockchain transactions;\n  - in `tests` we collect all reproducible Pytest tests that are executed in the scope of this Atomkraft project;\n  - finally, `reports` hold the test reports.\n- Configuration files:\n  - `atomkraft.toml` contains Atomkraft project configuration;\n  - `chain.toml` contains the blockchain configuration parameters, such as the number of nodes or the number of validators;\n  - `model.toml` contains the TLA+ model configuration parameters, used to run model checkers;\n  - `pyproject.toml` contains the Poetry project configuration.\n\nAll of the above configuration files are filled with the default configuration, so you don't need to worry about them, if you don't want to. On the other hand, when needed, everything is fully customizable.\n\n### Local testnets\n\nWith Atomkraft project created, you should be ready to go. By default, we configure local testnets to use the `gaiad` (Cosmos Hub) binary. If it is available via your `PATH`, executing `atomkraft chain testnet` should bring up a local testnet with 2 nodes and 3 validators. If you would like to configure any parameters differently (e.g. to run your custom blockchain binary), you can do it either via `atomkraft chain config` command, or by directly editing `chain.toml` config file. Please make sure your changes are valid by executing `atomkraft chain testnet`; we use the local testnet to run the tests.\n\n### Traces and reactors\n\nFor describing test cases we use the [_Informal Trace Format_](https://apalache.informal.systems/docs/adr/015adr-trace.html), which is a JSON encoding of a sequence of steps, and each step encodes values of key state variables; please see [this example trace](examples/cosmos-sdk/transfer/example_trace.itf.json). The trace has been produced by our in-house [Apalache model checker](https://apalache.informal.systems) from [this TLA+ model](examples/cosmos-sdk/transfer/transfer.tla).\n\nITF traces are abstract; in the example trace above, it holds wallet balances as a simple mapping `balances` from abstract addresses, represented as integers, to their balances, also represented as integers without denomination. There are two abstract actions in this trace: `Init`, and `Transfer`. In order to be able to replay the abstract trace on a blockchain, each of those abstract actions needs to be translated to a concrete transaction, as well as all abstract parameters of an action need to be translated to concrete values (e.g. an abstract integer address needs to be translated into a concrete Cosmos address). This translation step is performed by a component that we call `reactor`: a reactor is a centerpiece for an Atomkraft project, without which it can't function, similar to a nuclear reactor for the atomic power plant. You can see [this example reactor](examples/cosmos-sdk/transfer/reactor.py) that is able to playback the above trace on the blockchain.\n\nThis separation between abstract traces and reactors, which apply abstract traces to concrete blockchains, is the crucial aspect of Atomkraft infrastructure; here is why:\n\n- Abstract traces are more maintainable than concrete tests. Concrete tests need to be updated on every small change in encoding, API, etc.; we have seen PRs with thousands of LOC concerning only with updating manually written tests. With the separation into traces and reactors, traces won't need to be updated in most cases; only the relatively small reactors will need small tweaks.\n- Abstract traces can be generated by whatever means: from model checkers such as [Apalache](https://apalache.informal.systems), via fuzzing, PBT, BDD, or even manually. How the abstract trace is produced, doesn't matter; it still can be executed using Atomkraft's infrastructure.\n- It is much easier to understand the intent of a test case expressed as an abstract trace, because of absence of excessive details.\n- Abstract traces are lightweight in terms of storage, which allows us to generate and maintain thousands of them, covering many corner cases, at no extra cost.\n\nWe have automated the process of writing a reactor via `atomkraft reactor` command. A user needs only to supply the lists of actions, and of state variables, and the command will generate a reactor stub with a function for each action; what remains is only to fill the body of each such function.\n\n### Generating traces from TLA+ models\n\nAs explained above, abstract traces can be obtained by whatever means; we do not constrain the user in this respect. The most time efficient method, from our point of view, is to generate traces from formal models expressed in [TLA+](https://lamport.azurewebsites.net/tla/tla.html), the specification language designed by Leslie Lamport. For a gentle introduction to TLA+ you may use the Informal's [TLA+ Language Reference Manual](https://apalache.informal.systems/docs/lang/index.html) and [TLA+ Basics Tutorial](https://mbt.informal.systems/docs/tla_basics_tutorials/). While TLA+ may look scary for beginners, we can assure you that learning it will greatly improve your productivity when reasoning about (and testing!) both protocols and code.\n\nThe good news is that we have done a thorough work in making user's life as easy as possible when working with TLA+ models, and using them to generate abstract test traces. All heavy-lifting work wrt. TLA+ models is done by our [Apalache](https://apalache.informal.systems) model checker. The model checker itself is meant for expert users; Atomkraft tries its best to hide excessive complexity from its users and exposes only the most essential functionality for working with models. You can access this functionality via `atomkraft model` command, which provides you with the functions like listed in the screenshot below:\n\n![Atomkraft model](docs/images/atomkraft-model.png)\n\nThe most important command in the scope of test case generation is `atomkraft model sample`. E.g. the command below (assuming you are in the Atomkraft project)\n\n```sh\natomkraft model sample --model-path models/transfer.tla --traces-dir traces --examples Ex\n```\n\nwill generate an abstract trace from the [transfer.tla](examples/cosmos-sdk/transfer/transfer.tla) model, and store the generated trace in the `traces` directory of your Atomkraft project.\n\n### Running the tests against testnet\n\nLet's assume you've done all the steps outlined above:\n\n1. created a fresh Atomkraft project using `atomkraft init`\n2. configured a Cosmos-based blockchain of your choice using `atomkraft chain`\n3. created a reactor stub using `atomkraft reactor`, and populated it with code\n4. generated abstract traces from a TLA+ model using `atomkraft model sample`, or created abstract traces by some other means\n\nThen you are ready to go, and execute your tests! We provide two commands for doing that:\n\n- `atomkraft test trace` will accept an abstract trace and a reactor, spin the local testnet, and execute the trace against the testnet\n- `atomkraft test model` is a convenience shorthand that combines step 4 above (`atomkraft model sample`) with `atomkraft test trace`, and allows you to execute tests directly from a TLA+ model, sidestepping explicit trace generation.\n\nBoth of the above `atomkraft test` commands populate the `tests` directory of your project with Pytest-based tests; so executing `pytest` inside your Atomkraft project at any point in time will reproduce all of your tests. In fact, the complete Atomkraft project directory is ready at any point in time to be exported, and used as a Pytest project, for example for reproducing your tests in the CI.\n\n## What's next: Atomkraft's immediate future\n\nAtomkraft's functionality outlined above represents the tool MVP: please feel free to employ it in your projects, and let us know of your experience: we are always ready to assist!\n\nThere are many more features that are planned or are already being implemented; so stay tuned. Below is the preview of the future Atomkraft functionality:\n\n- **Standard test suites**: we have already started the effort to provide standard reactors and tests for most important Cosmos SDK modules: [bank](https://docs.cosmos.network/master/modules/bank/), [staking](https://docs.cosmos.network/master/modules/staking/), [authz](https://docs.cosmos.network/master/modules/authz/); you name it! This will serve the community at large, and will allow you, as an Atomkraft user, to easily bootstrap your new Atomkraft projects:\n  - running standard test suites in your CI will make sure that your new functionality doesn't break the important blockchain invariants;\n  - using standard test suites as blueprints will allow you to easily create your own tests suites via examination and adaptation of already existing tests.\n- **Test understanding and debugging**: we started to work in the direction of assisting the user in simplified creation and understanding of the test scenarios, as well as debugging the failed test runs:\n  - for the former, we will provide the differential trace viewer, which highlights only the changes between the trace steps;\n  - for the latter, we will provide the built-in blockchain explorer, and integrate into it the capabilities to trace back and forth between abstract trace steps and concrete blockchain transactions.\n- **Test reports and dashboards**: we plan to implement the functionality for generation of test reports as well as live dashboards that would provide an easy-to-grasp overview and categorization of executed and running tests.\n- **Exhaustiveness**: we plan to implement certain coverage metrics (e.g. transaction sequences up to specified length), and to help user achieving full coverage according to those metrics to provide confidence in the code in case no bugs are discovered.\n",
    'author': 'Andrey Kuprianov',
    'author_email': 'andrey@informal.systems',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/informalsystems/atomkraft',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import ast
import json
from os import PathLike
from typing import Set

from modelator.pytest.decorators import dict_get_keypath

from . import constants


def get_all_trace_actions(trace: PathLike, keypath: str) -> Set[str]:
    all_actions = set()
    with open(trace) as f_trace:
        itf_dict = json.load(f_trace)

    states = itf_dict["states"]
    for state in states:
        action = dict_get_keypath(state, keypath)
        all_actions.add(action)

    return all_actions


def get_keypath(root_node: ast.Module) -> str:
    for el in root_node.body:
        if (
            isinstance(el, ast.Assign)
            and len(el.targets) == 1
            and el.targets[0].id == constants.KEYPATH_VAR
        ):
            return el.value.value
    raise ValueError(
        f"Expression of the form {constants.KEYPATH_VAR} = <val> is missing from the reactor file."
    )



from typing import List, Optional

import typer

from .trace import test_trace

app = typer.Typer(rich_markup_mode="rich", add_completion=False)


def FileOption(help, default):
    return typer.Option(
        None,
        show_default=False,
        help=f"{help} [grey30][default: set via [bold cyan]atomkraft {default}[/bold cyan]][/grey30]",
    )


def RequiredFileOption(help, default):
    return typer.Option(
        ...,
        show_default=False,
        help=f"{help} [grey30][default: set via [bold cyan]atomkraft {default}[/bold cyan]][/grey30]",
    )


@app.command()
def trace(
    # currently, require the trace to be present.
    # later, there will be an option to pick up the last one from the model
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Output logging on console"
    ),
    trace: typer.FileText = RequiredFileOption("trace to execute", "model"),
    reactor: typer.FileText = FileOption("reactor to interpret the trace", "reactor"),
    keypath: str = typer.Option(
        "action",
        show_default=True,
        help="Path to key used as step name, extracted from ITF states",
    ),
):
    """
    Test blockchain by running one trace
    """

    test_trace(
        trace.name, reactor if reactor is None else reactor.name, keypath, verbose
    )


@app.command()
def model(
    model: typer.FileText = FileOption("model used to generate traces", "model"),
    config: typer.FileText = FileOption("model configuration", "model"),
    reactor: typer.FileText = FileOption("reactor to interpret the traces", "reactor"),
    test: Optional[List[str]] = typer.Option(
        None,
        show_default=False,
        help="model operator(s) describing test traces. multiple can be given either comma-separated, or via separate --test options",
    ),
):
    """
    Test blockchain by running multiple traces generated from a model
    """
    tests = [ts.split(",") for ts in test]
    tests = [t.strip() for ts in tests for t in ts]
    print(
        f"""
    Model: {model}
    Config: {config}
    Reactor: {reactor}
    Tests: {tests}
    """
    )
